
/* Clock Frequency. Change the clock frequency to the clock*/
/* frequency of your module. E.g. for a HERON1_6701 the fre*/
/* quency is 166666 kHz. You can also change the frequency */
/* to the frequency of any external clock.                 */

#define INPUT_FREQUENCY		166666

/***********************************************************/
/*                                                         */
/* This file implements the output_clock function. It pro- */
/* grams timer0 to output a clock signal to timer0 "tout"  */
/* pin. On HEPC8 systems this "tout" pin can be connected  */
/* to UMI lines (UMI=Uncommitted Module Interconnect), by  */
/* using the HeronUmiOut() function.                       */
/*                                                         */
/*     JT    ver 1.0  10/05/00   first written             */
/***********************************************************/

/* 
    Description of the bits in the control register:

	BIT     NAME            R/W     VALUE           DESCRIPTION
	 0      func            RW      0x001           1: timer pin, 0: general purpose i/o pin
	 1      i/o             RW      0x002           (used if bit0 is 0) 1: output pin, 0: input pin
	 2      datout			RW      0x004           (used if bit0 is 0 && bit1 is 1) value driven out on TIM pin
	 3      datin           R       0x008           (if bit0 is 0 && bit1 is 0) value on TIM pin
	 4      pwid            RW      0x010           Pulse width. 
	 5      reserved        R       0x020
	 6      go              RW      0x040           1: (if bit7 = 1) zero counter & start counting
	 7      hld             RW      0x080           1: counter allowed to count, 0: counter disabled
	 8      c/p             RW      0x100           1: clock mode, 0: pulse mode
	 9      clksrc			RW      0x200           1: internal clock f=cpu-clock/4, 0: value on TIM pin
	10      inv             RW      0x400           inverts the TINP pin signal 
	11      tstat           R       0x800           Timer status. Value of timer output.
*/

typedef union timer_control
{
	unsigned int c;
	struct
	{
		unsigned int func:1;			/* bit  0 */
		unsigned int invout:1;			/* bit  1 */
		unsigned int datout:1;			/* bit  2 */
		unsigned int datin:1;			/* bit  3 */
		unsigned int pwid:1;			/* bit  4 */
		unsigned int rsvd:1;			/* bit  5 */
		unsigned int go:1;				/* bit  6 */
		unsigned int hld:1;				/* bit  7 */
		unsigned int cp:1;				/* bit  8 */
		unsigned int clksrc:1;			/* bit  9 */
		unsigned int invinp:1;			/* bit 10 */
		unsigned int tstat:1;			/* bit 11 */
	} x;

} CONTROL;

int output_clock(int freq)	/* freq in kHz!!! */
{
   volatile unsigned int *timer0_control = (unsigned int *) 0x01940000;
   volatile unsigned int *timer0_period  = (unsigned int *) 0x01940004;

	CONTROL      control;
	unsigned int cycles;

	control.c        = 0;

	control.x.func   = 1;		/* TOUT is a timer output pin   */
	control.x.invout = 0;		/* Uninverted TSTAT drives TOUT */
	control.x.datout = 0;		/* DATOUT is driven onto TOUT   */
	control.x.datin  = 0;		/* R                            */

	control.x.pwid   = 0;		/* Puls width is 1 clock cycle  */
	control.x.rsvd   = 0;		/* Reserved                     */
	control.x.go     = 1;		/* Reset timer & start counting */
	control.x.hld    = 1;		/* Enable counter               */

	control.x.cp     = 1;		/* TSTAT has 50% duty cycle     */
	control.x.clksrc = 1;		/* Internal clock: Hz/4         */
	control.x.invinp = 0;		/* Irrelevant                   */

	/* Assume 167 Mhz clock. There is 1 tick per 4 internal clock cycles */
	cycles = (INPUT_FREQUENCY/(freq*8));

	/* External clock, don't invert TINP pin */
	*timer0_control = (unsigned int) 0x00000001;	/* Halt the timer              */
	*timer0_period  = (unsigned int) cycles;		/* Program the period register */
	*timer0_control = (unsigned int) control.c;		/* GO!                         */

	return cycles;
}
