/******************************************************************/
/*                                                                */
/*  Flash ROM program for HERON4                                  */
/*                                                                */
/*  This program will either load a *.out file out of flash       */
/*  memory, or if there's no program loaded, try to load a        */
/*  *.out file from FIFO 0. Meanwhile, via interrupts, it         */
/*  will try to reply to I2C requests.                            */
/*                                                                */
/*	JT		ver 1.0		20/07/00	Newly Created                 */
/*	JT		ver 2.0		20/07/00	I2C support added             */
/*  JT      ver 3.0     08/08/00    Protocol changed. Use ints    */
/*                                                                */
/******************************************************************/

#include "i2c.h"

/* Version */
#define VERSION			3

/************************************************************/

#define TOUT_IS_TIMER 0x1
#define TOUT_IS_OUTPUT 0x0

#define INVERT_TIMER_OUTPUT 0x02
#define DONT_INVERT_TIMER_OUTPUT 0x00

#define DATAOUT_HIGH_WHEN_OUTPUT TIMER 0x4
#define DATAOUT_LOW_WHEN_OUTPUT TIMER 0x0

#define DATAIN 0x8

#define TOUT_PULSE_WIDTH_IN_PULSE_MODE_IS_2_CYCLES 0x10
#define TOUT_PULSE_WIDTH_IN_PULSE_MODE_IS_1_CYCLE 0

#define STOP_TIMER 0x00
#define RESUME_COUNTING 0x80
#define START_FROM_ZERO 0xc0

#define PULSE_MODE 0
#define DUTY_CYCLE_50_PERCENT 0x100

#define CLOCK_SOURCE_QUARTER_CLKOUT1 0x200
#define CLOCK_SOURCE_TINP 0

#define INVERT_TINP_FOR_CLOCK_SOURCE 0x400
#define DONT_INVERT_TINP_FOR_CLOCK_SOURCE 0

#define TSTAT 0x800
#define CLKOUT1_RATE 132000000

#define REFPER (int)((double)CLKOUT1_RATE * 15.625E-6 / 2.0)

extern cregister volatile unsigned int AMR;
extern cregister volatile unsigned int CSR;
extern cregister volatile unsigned int IFR;
extern cregister volatile unsigned int ISR;
extern cregister volatile unsigned int ICR;
extern cregister volatile unsigned int IER;
extern cregister volatile unsigned int ISTP;
extern cregister volatile unsigned int IRP;
extern cregister volatile unsigned int NRP;
extern cregister volatile unsigned int IN;
extern cregister volatile unsigned int OUT;

/*
   I2C interrupt
*/

volatile unsigned int *i2c_data  = (volatile unsigned int *)0x016C0000;
volatile unsigned int *i2c_ctrl  = (volatile unsigned int *)0x016D0000;
volatile unsigned int *i2c_s_ad  = (volatile unsigned int *)0x016E0000;
volatile unsigned int *i2c_m_ad  = (volatile unsigned int *)0x016F0000;

volatile unsigned int *dsp_inten = (volatile unsigned int *)0x01650000;
volatile unsigned int *hpic      = (volatile unsigned int *)0x01880000;

volatile unsigned int i2c_ctrl_mask = 0;
volatile unsigned int slave_id;

void Initialise_I2C_Device(int modid);
int  Receive_I2C_Message  (void);
int  Process_I2C_Message  (unsigned int count);
int  Send_I2C_Message     (unsigned int count);

interrupt void HostPortIsr(void)
{
   volatile unsigned int flags, s;

   flags = *i2c_ctrl;

   if ((flags & 0x01) != 0)
   {
      s = Receive_I2C_Message( );
      s = Process_I2C_Message(s);
      if (s) Send_I2C_Message(s);
   }

   /* Clear the DSPINT flag in the HPIC register */
   *hpic |= 0x20000;             
}

/*
   Timer interrupt
*/

#define FIFO0_EF_MASK 0x1000  /* Bit 12 = 1: Can read one              */

int readfifo = 0;

interrupt void tick(void)
{
   volatile int *in_flags = (int *)0x01640000;
#if XMAS
   volatile int *dig_io   = (int *)0x01660000;
   static   int  flash    = 0;

   if (flash) *dig_io = ~VERSION;
   else       *dig_io =  0xff;
   flash = !flash;
#endif

   if (*in_flags & FIFO0_EF_MASK)
   {
	   readfifo = 1;
   }
}

/*
   Copy memory using DMA functions
*/

void dma_copy(unsigned int from, unsigned int to, int n)
{
   volatile unsigned int *dma_ch0_primary_control     = (unsigned int *) 0x01840000;
   volatile unsigned int *dma_ch0_secondary_control   = (unsigned int *) 0x01840008;
   volatile unsigned int *dma_ch0_source_address      = (unsigned int *) 0x01840010;
   volatile unsigned int *dma_ch0_destination_address = (unsigned int *) 0x01840018;
   volatile unsigned int *dma_ch0_transfer_counter    = (unsigned int *) 0x01840020;

   *dma_ch0_primary_control     = 0;                /* Stop/initialise DMA */

   *dma_ch0_source_address      = from;
   *dma_ch0_destination_address = to;
   *dma_ch0_transfer_counter    = n;
   *dma_ch0_secondary_control   = 0x00000000;
   *dma_ch0_primary_control     = 0x01000251;           /* start dma */

   while ((*dma_ch0_primary_control&0xc)!=0) {}
}

/* 
   Read from FIFO
*/

int read_a_word()
{
   int           data_in;
   volatile int *fifo0    = (int *)0x00400000;
   volatile int *in_flags = (int *)0x01640000;

   while ((*in_flags & FIFO0_EF_MASK) == 0)
   {
		asm("	nop 5");
   }

   data_in = *fifo0;

   return (data_in);
}

int entry_point;

int boot_program(int modid)
{
   volatile int *dma0_primary_control  = (int *)0x01840000;
   volatile int *dma0_secondary_control= (int *)0x01840008;
   volatile int *dma0_source_addr      = (int *)0x01840010;
   volatile int *dma0_dest_addr        = (int *)0x01840018;
   volatile int *dma0_transfer_count   = (int *)0x01840020;
   volatile int *dma_cnt_reload_reg_A  = (int *)0x01840028;
   volatile int *dig_io                = (int *)0x01660000;

   int          *block_addr;
   int          i,block_size,read_but_ignore;
   int          temp;
   volatile int mod_car;
   volatile int dummy;

   read_but_ignore = 1;

   mod_car = read_a_word() & 0xff;
   if (mod_car == modid)
   {
      read_but_ignore = 0;			/* it's for us                               */
      IER            &= 0x9FFF;		/* Disable timer & host port (I2C) interrupt */
	  CSR			 &= (~1);		/* Globally disable interrupts               */
	  ICR		      = 0xfff0;		/* Clear interrupts 4..15    */
   }
   else
   {
      read_but_ignore = 1;            /* it's not for us */
   }

   block_size = read_a_word();
   block_size = block_size >> 2;
   while (block_size != 0)
   {
      block_addr = (int *)read_a_word();
      if (read_but_ignore)	/* got to read it, but don't write */
	  {					/* to memory.                      */
         for (i=0;i<block_size;i++)
		 {
            dummy = read_a_word();
		 }
	  }
      else
	  {							/* it's for us! */   
         if ((int)block_addr < 0x00010000)	/* If it's internal program RAM */
		 {						/* we have to use single word DMA */
            for(i=0;i<block_size;i++)
			{
               temp                    = read_a_word();
               *dma0_secondary_control = 0x0000a000;
               *dma0_source_addr       = (int)(&temp);
               *dma0_dest_addr         = (int)(block_addr+i);
               *dma0_transfer_count    = 1;
               *dma_cnt_reload_reg_A   = 1;
               *dma0_primary_control   = 0x01000051;
               while((*dma0_primary_control & 0x0000000c) != 0)
			   {
                  ;
			   }
			}
		 }
         else						/* otherwise we can use regular */
		 {						/* CPU data writes to load the code */
            for(i=0;i<block_size;i++)
			{
               *block_addr++ = read_a_word();
			}
		 }
	  }

      block_size = read_a_word();
      block_size = block_size >> 2;
   }
   entry_point = read_a_word();                /* Last word is the entry point */
   *dig_io     = 0xf;

   return (read_but_ignore);
}

void Init_Timer(void);

main()
{
   volatile unsigned int *flash;
   volatile unsigned int  address, size, magic;

   volatile int      *mod_car_id       = (int *)0x01670000;
   volatile int        *emif_gcr       = (int *)0x01800000;
   volatile int *emif_ce1_timing       = (int *)0x01800004;
   volatile int *emif_ce0_timing       = (int *)0x01800008;
   volatile int *emif_ce2_timing       = (int *)0x01800010;
   volatile int *emif_ce3_timing       = (int *)0x01800014;
   volatile int *emif_sdram_control    = (int *)0x01800018;
   volatile int *emif_sdram_refresh    = (int *)0x0180001c;
   volatile int *dig_io                = (int *)0x01660000;

   int          j, modid;

   /* Program the EMIF */
   /* For 120Mhz clocked HERON4 (:HEPC8) */
   *emif_gcr           = 0x000030fc;	/* HERON1: 0x000030f8; */
   *emif_ce0_timing    = 0x00000030;	/* HERON1: 0x00000040; */
   *emif_ce1_timing    = 0x34b12823;	/* HERON1: 0x34b1e803; */
   *emif_ce2_timing    = 0x00000030;	/* HERON1: 0x12020920; */
   *emif_ce3_timing    = 0x00000040;	/* HERON1: 0x00000030; */
   *emif_sdram_control = 0x07114000;	/* HERON1: 0x07229000; */
   *emif_sdram_refresh = REFPER;		/* 3A9:  120MHz 'c6701 */

   /* =========================================================== */
   /* First look in flash memory and see if there's a boot stream */
   flash   = (unsigned int *) 0x1402000;
   address = *flash++;
   size    = *flash++;
   magic   = *flash++;
   j       =  1;
   while ((size)&&(magic==0xdeafd00f))
   {
      *dig_io = ~j++;
	  /* Copy this segment to memory */
      dma_copy((unsigned int) flash, address, size);
      flash  += ((size+3)>>2);          /* size is in bytes */
      address = *flash++;
      size    = *flash++;
      magic   = *flash++;
   }

   if (magic==0xdeafd00f)
   {
      *dig_io = 0xf;
      /* We reached the end of the boot stream in flash memory */
      /* The last entry should be the entry point. Read it.    */
      entry_point = *flash++;
      goto jump2userprogram;
   }
   /* ======================================================== */

   /* ========================================================== */
   /* If we're here we didn't find a proper boot stream in flash */
   /* No problem, just wait for the FIFO to find a boot stream.  */
   /* But, first initialise the I2C bus to accept probes         */

   *dig_io = ~VERSION;				/* Indicate our version 2    */
   modid   = ((*mod_car_id)&0xff);
   ICR     = 0xfff0;				/* Clear interrupts 4..15    */
   Initialise_I2C_Device(modid);	/* Prepare I2C interrupts    */
   Init_Timer();					/* Prepare timer interrupts  */
   CSR     = CSR|1;					/* Globally enable interrupts*/

   j = 1;
   while (j)
   {
	  if (readfifo)
	  {
         readfifo = 0;
         j        = boot_program(modid);
	  }
      asm("   nop     5");
   }

jump2userprogram:

   asm("   nop     5");
   asm("   ldw     *+b14(_entry_point),a0");
   asm("   nop     5");
   asm("   b.s2    a0");   /* Branch to user program */
   asm("   nop     5");
}

#define MAX		16
unsigned char buffer[MAX];

int Receive_I2C_Message(void)
{
   volatile          int i;
   volatile unsigned int status,trash;

   i = 0;
   do
   {
      /* Wait for a data byte to be received */
      /* Poll the empty flag */
      do
	  {
         status  = *i2c_ctrl;
         status &= 0x05;
//         asm("	nop 5");
	  }
      while (status == 0x04);
	
      if (status == 0x05)
	  {
         if (i<MAX)
		 {
            buffer[i] = (*i2c_data & 0xff);
		 }
         else
		 {
            trash = *i2c_data;
		 }
         i++;		
	  }
   }
   while (status == 0x05);

   /* Wait for the I2C device to return to the correct state */
   do
   {
      status = (*i2c_ctrl & 0x1f);
//      asm("	nop 5");
   }
   while (status != 0x12);

   return (i);
}

int Process_I2C_Message(unsigned int count)
{
	if (count > 1)
	{
		switch (buffer[0])
		{
		case QUERY_TYPE :
			buffer[0] = buffer[1];				/* return address     */
			buffer[1] = RESP_TYPE;				/* response type      */
			buffer[2] = slave_id;				/* sender I2C address */
			buffer[3] = PROCESSOR;				/* module type        */
			buffer[4] = FAMILY;					/* HERON type         */
			buffer[5] = PROCTYPE;				/* Processor type     */
			buffer[6] = VERSION;				/* ROM version		  */
			buffer[7] = 0;						/* Option number      */
			return 8;
		}
	}
	return 0;
}

int Send_I2C_Message(unsigned int count)
{
   volatile unsigned int status;
   volatile int i;

   /* Set up the master address */
   *i2c_m_ad      = (buffer[0]<<1);

   /* Set the Send Message bit */
   i2c_ctrl_mask |= 0x20;
   *i2c_ctrl      = i2c_ctrl_mask;

   /* message is from i=1 onwards */
   i = 1;
   while (i<count)
   {
      /* Send a data byte */
      *i2c_data = buffer[i];

      do
	  {
         status  = *i2c_ctrl;
         status &= 0x12;
//         asm("	nop 5");
	  }
      while (status == 0x10);
    
      if (status != 0x12) goto eom;
	  i++;
   }

   status = ((*i2c_ctrl) & 0x1f);
   if ((status & 0x10) == 0)
   {
      /* Wait for the End of Message bit to be set */
      while ((status & 0x08) == 0)
	  {
//         asm("	nop 5");
         status = ((*i2c_ctrl) & 0x1f);
	  }
   }

eom:
   /* Clear the send message bit */
   i2c_ctrl_mask &= 0x1f;
   *i2c_ctrl      = i2c_ctrl_mask;

   /* Wait for the I2C device to return to the correct state */
   do
   {
      status = (*i2c_ctrl & 0x1f);
//      asm("	nop 5");
   }
   while (status != 0x12);

   return 1;
}

void Initialise_I2C_Device(int modid)
{
	volatile unsigned int *ptr;

   /*****************************************************/
   /* Set up the slave address, as per Pete's document: */
   /* bits 012 form the I2C address:                    */
   /*        0       = reserved                         */
   /*        1/2/3/4 = HERON id                         */
   /*        5       = host bus i2c node                */
   /*        6       = interboard connection node       */
   /*        7       = HEART FPGA                       */
   /* bits 3456 form the board is as set by the switch  */
   /*****************************************************/
   slave_id      =  (modid&0x07);
   slave_id     |= ((modid&0xf0)>>1);

   *i2c_s_ad     =  (slave_id<<1);
   *i2c_ctrl     =   0;

   /* Enable I2C Interrupts (==HPI interrupt) */
   IER          |= 0x2002;
   /* Clear the DSPINT flag in the HPIC register */
   *hpic        |= 0x20000;             

   /* Enable I2C Interrupts on the HERON4 */
   i2c_ctrl_mask = 0x05; 
   *i2c_ctrl     = i2c_ctrl_mask;

   /* Keep I2C data also in a local memory area at fixed  */
   /* address, so that we can find the data whatever the  */
   /* heron module type is. Used by plug-ins, who have a  */
   /* hard time doing i2c.								  */
   ptr = (volatile unsigned int *) DATAADDR;
   ptr[0] = MAGIC;					/* verification	      */
   ptr[1] = RESP_TYPE;				/* response type      */
   ptr[2] = slave_id;				/* sender I2C address */
   ptr[3] = PROCESSOR;				/* module type        */
   ptr[4] = FAMILY;					/* HERON type         */
   ptr[5] = PROCTYPE;				/* Processor type     */
   ptr[6] = VERSION;				/* ROM version		  */
   ptr[7] = 0;						/* Option number      */
}

void Init_Timer(void)
{
   volatile int *timer0_control        = (int *)0x01940000;
   volatile int *timer0_period         = (int *)0x01940004;

   /* ================= Tick interrupt so we can see it's running */
   *timer0_control = TOUT_IS_TIMER | STOP_TIMER      | CLOCK_SOURCE_QUARTER_CLKOUT1;
   *timer0_period  = (CLKOUT1_RATE >> 2) * 0.1;
   *timer0_control = TOUT_IS_TIMER | START_FROM_ZERO | CLOCK_SOURCE_QUARTER_CLKOUT1;

   /* Interrupt clear. (4 through to 15)        */
   /* Set GIE bit (Global Interrupt Enable bit) */
   IER |= 0x4002;		/* j   = (1<<1) | (1<<14); */
}


/* This is the Interrupt Service Fetch Packet for RESET.                */
/* It must be programmed at 0 so that when the DMA has finished,        */
/* the first thing that executes is the branch this program.            */

asm("   .global _c_int00");
asm("   .sect   .vectors");

asm("l0:        NOP");
asm("   NOP     ");
asm("   b.s2    _c_int00");
asm("   nop     ");
asm("   nop     ");
asm("   nop     ");
asm("   nop     ");
asm("   nop     ");

asm("   .space  32");	/* l1  */
asm("   .space  32");	/* l2  */
asm("   .space  32");	/* l3  */
asm("   .space  32");	/* l4  */
asm("   .space  32");	/* l5  */
asm("   .space  32");	/* l6  */
asm("   .space  32");	/* l7  */
asm("   .space  32");	/* l8  */
asm("   .space  32");	/* l9  */
asm("   .space  32");	/* l10 */
asm("   .space  32");	/* l11 */
asm("   .space  32");	/* l12 */

asm("l13:        NOP");
asm("   NOP     ");
asm("   b.s2    _HostPortIsr");
asm("   nop     ");
asm("   nop     ");
asm("   nop     ");
asm("   nop     ");
asm("   nop     ");

#if 1
asm("l14:        NOP");
asm("   NOP     ");
asm("   b.s2    _tick");
asm("   nop     ");
asm("   nop     ");
asm("   nop     ");
asm("   nop     ");
asm("   nop     ");

asm("l15:        NOP");
asm("   NOP     ");
asm("   b.s2    _tick");
asm("   nop     ");
asm("   nop     ");
asm("   nop     ");
asm("   nop     ");
asm("   nop     ");
#else
asm("   .space  32");	/* l14 */
asm("   .space  32");	/* l15 */
#endif
